//-
// Copyright 2017, 2018 The proptest developers
//
// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
// http://www.apache.org/licenses/LICENSE-2.0> or the MIT license
// <LICENSE-MIT or http://opensource.org/licenses/MIT>, at your
// option. This file may not be copied, modified, or distributed
// except according to those terms.

//! Arbitrary implementations for libcore.

mod ascii;
mod cell;
mod cmp;
mod convert;
mod fmt;
mod iter;
mod marker;
mod mem;
mod num;
mod option;
mod result;
